/* Meine eigene Version von ping.
 *
 * Copyright (c) 2002
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * NEUE VERSION! Ich hab den Tipp bekommen, dass die ICMP-Checksumme
 * nicht gebildet wird und daher einiges Hosts nicht geantwortet haben.
 * Das habe ich jetzt nachgeholt, und siehe da: es funktioniert.
 *
 * Es hat zwar nicht ganz so viele Funktionen wie das "normale" ping,
 * aber es funktioniert und bietet einen einfachen Einblick in die
 * Programmierung mit RAW-Sockets und ICMP. Fehler oder Probleme bitte
 * an felix@zotteljedi.de mailen.
 *
 * Dieses Programm funktioniert in der Form nur unter Linux (aufgrund 
 * der Benennung der structs und deren Elemente - unter BSD sind sie
 * anders. Sollte aber nicht schwer fallen das Programm zu portieren)
 *
 *
 * Datum       nderung
 * ---------------------------------------------------------------------------
 * 2002-05-11  Berechnung der Checksumme korrigiert.
 * 2000-??-??  Ursprngliche Implementierung.
 */

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netinet/ip_icmp.h>

#define BUF_SIZ 128

int seq, s, n_recvd, empty_len, count = -1, timeout = 1;
pid_t pid;
struct sockaddr_in to;
const int data_len = sizeof(struct timeval);
const int icmp_len = sizeof(struct icmphdr);
char *hostname;
double rtt_min, rtt_avg, rtt_max;

u_short in_cksum(u_short *addr, int len)
{
	u_short i = 0, *word = addr;
	u_long acc = 0;
				
	while(i++ < len / 2)
		acc += *(word++);
					
	return ~(*(u_short*)&acc + *((u_short*)&acc + 1));
}


void send_packet(int sig)
{
	char buffer[BUF_SIZ];
	void *data;
	struct icmphdr *icmp;
	struct timeval tv;
	int bytes;
	
	if(sig != SIGALRM)
		return;
	
	icmp = (struct icmphdr*) buffer;
	bzero(icmp, sizeof(struct icmphdr));
	
	icmp->type = ICMP_ECHO;
	icmp->code = 0;
	icmp->un.echo.id = pid;
	icmp->un.echo.sequence = seq;
	
	data = buffer + sizeof(struct icmphdr);
	
	if (gettimeofday(&tv, NULL) == -1)
	{
		perror("gettimeofday() failed");
		exit(4);
	}
	
	memcpy(data, &tv, sizeof(struct timeval));

	icmp->checksum = 0;
	icmp->checksum = in_cksum((u_short*)icmp, sizeof(struct timeval) + sizeof(struct icmphdr));
	
	bytes = sendto(s, buffer, icmp_len + data_len + empty_len, 0, &to, sizeof(to));
	if (bytes == -1)
	{
		perror("sendto() failed");
		exit(3);
	}
	
	seq++;
	
	alarm(timeout);
}

void time_diff(struct timeval *recvd, struct timeval *sent)
{
	recvd->tv_usec -= sent->tv_usec;
	if (recvd->tv_usec < 0)
	{
		recvd->tv_usec += 1000000;
		recvd->tv_sec--;
	}
	
	recvd->tv_sec -= sent->tv_sec;
}

void got_packet(char *buf, int buf_len, struct timeval *recvd)
{
	struct iphdr *ip;
	struct icmphdr *icmp;
	struct in_addr addr;
	double diff;
	void *data;
	
	ip = (struct iphdr*) buf;
	icmp = (struct icmphdr*) (buf + sizeof(struct iphdr));
	data = buf + sizeof(struct iphdr) + sizeof(struct icmphdr);
	
	if (icmp->un.echo.id != pid)
		return;
	
	n_recvd++;
	
	addr.s_addr = ip->saddr;
	time_diff(recvd, data);
	diff = recvd->tv_sec * 1000.0 + recvd->tv_usec / 1000.0;
	
	printf("%i bytes from %s: seq=%i ttl=%i time=%.3f ms\n",
		buf_len, inet_ntoa(addr), icmp->un.echo.sequence, ip->ttl, diff);
		
	if (diff > rtt_max)
		rtt_max = diff;
	
	if (diff < rtt_min)
		rtt_min = diff;
	
	if (rtt_min == 0.0)
		rtt_min = diff;
		
	rtt_avg += diff;
}

void view_summary(int sig)
{
	double loss, tmp;
	
	if (sig != SIGINT)
		return;
	
	tmp = 100.0 / seq;
	loss = tmp * (seq - n_recvd);
	
	printf("--- statistics for %s ---\n", hostname);
	
	printf("%i packets sent, %i packets received, %i%% packet loss\n", seq, n_recvd, (int)loss);
	
	printf("round-trip min/avg/max = %.3f/%.3f/%.3f ms\n",
		rtt_min, rtt_avg/n_recvd, rtt_max);
		
	exit(0);
}

int get_options(int argc, char *argv[])
{
	int i;
	for (i = 0; i < argc; i++)
	{
		if(!strcmp(argv[i], "-s") && (argc - 1 > i))
		{
			empty_len = atol(argv[i + 1]) - data_len;
			if (empty_len < 0)
				empty_len = 0;
		}
		
		if(!strcmp(argv[i], "-i") && (argc - 1 > i))
		{
			timeout = atol(argv[i + 1]);
			if (timeout < 1)
				timeout = 1;
				
		}
		
		if(!strcmp(argv[i], "-c") && (argc - 1 > i))
		{
			if (atol(argv[i + 1]) > 0)
				count = atol(argv[i + 1]);
		}
		
		if(!strcmp(argv[i], "--info"))
		{
			printf("my version of ping - written by Felix Opatz\n");
			printf("please mail bugreports to gallenstein@web.de\n");
			exit(0);
		}
		
	}
	
	return 0;
}

int main(int argc, char *argv[])
{
	char buffer[BUF_SIZ];
	int bytes;
	struct timeval now;
	struct hostent *host;
	
	if (argc < 2)
	{
		printf("%s <host> [-s packetsize] [-i wait] [-c count] [--info]\n", argv[0]);
		return 1;
	}
	
	hostname = argv[1];
	
	get_options(argc, argv);
	
	s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (s == -1)
	{
		perror("socket() failed");
		return 2;
	}
	
	seteuid(getuid());
	pid = getpid();
	
	host = gethostbyname(hostname);
	if (host == NULL)
	{
		herror("gethostbyname() failed");
		return 3;
	}
	
	to.sin_addr = *(struct in_addr*)host->h_addr;
	printf("ping %s (%s): %i data bytes\n",
		host->h_name, inet_ntoa(to.sin_addr), data_len + empty_len);
	
	to.sin_family = AF_INET;
	to.sin_port = 0;
	
	signal(SIGALRM, send_packet);
	signal(SIGINT, view_summary);
	send_packet(SIGALRM);
	
	for(;;)
	{
		bytes = recv(s, buffer, sizeof(buffer), 0);
		if (bytes == -1)
		{
			if(errno == EINTR)
				continue;
			
			perror("recv() failed");
			return 4;
		}
		
		gettimeofday(&now, NULL);
		got_packet(buffer, bytes, &now);
		
		if (seq == count)
			kill(getpid(), SIGINT);
	}
		
	return 0;
}

